import glob
import json
import os
import signal
import subprocess
import sys
import textwrap
import logging
import logging.config
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from functools import reduce
from itertools import groupby
from pathlib import Path
from threading import Thread


def create_task_logger(task_control_dir):
    if os.environ.get("DRYPIPE_TASK_DEBUG") != "True":
        logging_level = logging.INFO
    else:
        logging_level = logging.DEBUG
    _logger = logging.Logger("task-logger", logging_level)
    file_handler = logging.FileHandler(filename=os.path.join(task_control_dir, "drypipe.log"))
    file_handler.setLevel(logging_level)
    file_handler.setFormatter(
        logging.Formatter("%(asctime)s - %(levelname)s - %(message)s", datefmt='%Y-%m-%d %H:%M:%S%z')
    )
    _logger.addHandler(file_handler)
    return _logger


__script_location = os.environ.get("__script_location")


if __script_location is None:
    logger = logging.getLogger(__name__)
else:
    logger = create_task_logger(__script_location)


class PortablePopen:

    def __init__(self, process_args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=None, shell=False, stdin=None):

        if stdout is None:
            raise Exception(f"stdout can't be None")

        if stderr is None:
            raise Exception(f"stderr can't be None")

        self.process_args = [str(p) for p in process_args]
        self.popen = subprocess.Popen(
            process_args,
            stdout=stdout,
            stderr=stderr,
            shell=shell,
            env=env,
            stdin=stdin
        )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.popen.__exit__(exc_type, exc_val, exc_tb)

    def wait(self, timeout=None):
        return self.popen.wait(timeout)

    def stderr_as_string(self):
        return self.popen.stderr.read().decode("utf8")

    def stdout_as_string(self):
        return self.popen.stdout.read().decode("utf8")

    def read_stdout_lines(self):
        for line in self.popen.stdout.readlines():
            yield line.decode("utf8")

    def safe_stderr_as_string(self):
        try:
            return self.stderr_as_string()
        except Exception as _:
            return ""

    def raise_if_non_zero(self):
        r = self.popen.returncode
        if r != 0:
            invocation = ' '.join(self.process_args)
            raise Exception(
                f"process invocation returned non zero {r}: {invocation}\nstderr: {self.safe_stderr_as_string()}"
            )

    def wait_and_raise_if_non_zero(self):
        self.popen.wait()
        self.raise_if_non_zero()


def python_shebang():
    return "#!/usr/bin/env python3"


def write_pipeline_lib_script(file_handle):
    file_handle.write(
        f"{python_shebang()}\n" + textwrap.dedent(f"""            
            import os
            import sys 
            import importlib.machinery
            import importlib.util        

            __script_location = os.path.dirname(os.path.abspath(__file__))
            script_lib_path = os.path.join(__script_location, 'script_lib.py')        
            loader = importlib.machinery.SourceFileLoader('script_lib', script_lib_path)
            spec = importlib.util.spec_from_loader(loader.name, loader)
            script_lib = importlib.util.module_from_spec(spec)
            loader.exec_module(script_lib)
        """)
    )

    file_handle.write(textwrap.dedent(f"""        
    if __name__ == '__main__':        
        script_lib.handle_script_lib_main()
    """))


def task_script_header():

    return f"{python_shebang()}\n" + textwrap.dedent(f"""            
        import os
        import sys 
        import importlib.machinery
        import importlib.util        

        is_slurm = os.environ.get("__is_slurm") == "True"        
        if is_slurm:
            __script_location = os.environ['__script_location']
        else:
            __script_location = os.path.dirname(os.path.abspath(__file__))
            os.environ["__script_location"] = __script_location
        script_lib_path = os.path.join(os.path.dirname(__script_location), 'script_lib.py')        
        loader = importlib.machinery.SourceFileLoader('script_lib', script_lib_path)
        spec = importlib.util.spec_from_loader(loader.name, loader)
        script_lib = importlib.util.module_from_spec(spec)
        loader.exec_module(script_lib)        
    """)


def _exit_process():
    _delete_pid_and_slurm_job_id()
    logging.shutdown()
    os._exit(0)


def run_python(python_bin, mod_func, container=None):
    switches = "-u"
    cmd = [
        python_bin,
        switches,
        "-m",
        "dry_pipe.cli",
        "call",
        mod_func
    ]

    if container is not None:
        cmd = [
            "singularity",
            "exec",
            os.path.join(os.environ['__containers_dir'], container)
        ] + cmd

    env = {**os.environ}

    if os.environ.get("__is_slurm"):
        env['__scratch_dir'] = os.environ['SLURM_TMPDIR']

    has_failed = False

    logger.info("run_python: %s", ' '.join(cmd))

    with open(os.environ['__out_log'], 'a') as out:
        with open(os.environ['__err_log'], 'a') as err:
            with PortablePopen(cmd, stdout=out, stderr=err) as p:
                try:
                    p.wait()
                    has_failed = p.popen.returncode != 0
                except Exception as ex:
                    has_failed = True
                    logger.exception(ex)
                finally:
                    if has_failed:
                        step_number, control_dir, state_file, state_name = read_task_state()
                        _transition_state_file(state_file, "failed", step_number)
    if has_failed:
        _exit_process()


def ensure_upstream_tasks_completed(env):
    non_completed_dependent_task = env.get("__non_completed_dependent_task")
    if non_completed_dependent_task is not None:
        print("upstream dependent task : " + non_completed_dependent_task + " not completed, cannot run.")


def _fail_safe_stderr(process):
    try:
        err = process.stderr.read().decode("utf-8")
    except Exception as _e:
        err = ""
    return err


def _terminate_descendants_and_exit(p1, p2):

    try:
        try:
            logger.info("signal SIGTERM received, will transition to killed and terminate descendants")
            step_number, control_dir, state_file, state_name = read_task_state()
            _transition_state_file(state_file, "killed", step_number)
            logger.info("will terminate descendants")
        except Exception as _:
            pass

        this_pid = str(os.getpid())
        with PortablePopen(
            ['ps', '-opid', '--no-headers', '--ppid', this_pid]
        ) as p:
            p.wait_and_raise_if_non_zero()
            pids = [
                int(line.decode("utf-8").strip())
                for line in p.popen.stdout.readlines()
            ]
            pids = [pid for pid in pids if pid != p.popen.pid]
            logger.debug("descendants of %s: %s", this_pid, pids)
            for pid in pids:
                try:
                    os.kill(pid, signal.SIGTERM)
                except Exception as _:
                    pass
    except Exception as ex:
        logger.exception(ex)
    finally:
        _exit_process()


def _delete_pid_and_slurm_job_id():
    try:
        sloc = os.environ["__script_location"]

        def delete_if_exists(f):
            f = os.path.join(sloc, f)
            if os.path.exists(f):
                os.remove(f)

        delete_if_exists("pid")
        delete_if_exists("slurm_job_id")
    except Exception as ex:
        logger.exception(ex)


def env_from_sourcing(env_file):

    p = os.path.abspath(sys.executable)

    dump_with_python_script = f'{p} -c "import os, json; print(json.dumps(dict(os.environ)))"'

    logger.debug("will source file: %s", env_file)

    with PortablePopen(['/bin/bash', '-c', f". {env_file} && {dump_with_python_script}"]) as p:
        p.wait_and_raise_if_non_zero()
        out = p.stdout_as_string()
        return json.loads(out)


def iterate_out_vars_from(file):
    if os.path.exists(file):
        with open(file) as f:
            for line in f.readlines():
                var_name, value = line.split("=")
                yield var_name.strip(), value.strip()


def source_task_env(task_env_file):

    env = env_from_sourcing(task_env_file)

    for k, v in env.items():
        os.environ[k] = v

    return env


def parse_in_out_meta(name_to_meta_dict):

    # export __meta_<var-name>="(int|str|float):<producing-task-key>:<name_in_producing_task>"
    # export __meta_<file-name>="file:<producing-task-key>:<name_in_producing_task>"

    #Note: when producing-task-key == "", meta are output values and files

    def gen():
        for k, v in name_to_meta_dict.items():
            var_name = k[7:]
            typez, task_key, name_in_producing_task = v.split(":")

            yield task_key, name_in_producing_task, var_name, typez

    def k(t):
        return t[0]

    for producing_task_key, produced_files_or_vars_meta in groupby(sorted(gen(), key=k), key=k):

        def filter_and_map(condition_on_typez):
            return [
                (name_in_producing_task, var_name, typez)
                for task_key, name_in_producing_task, var_name, typez
                in produced_files_or_vars_meta
                if condition_on_typez(typez)
            ]

        yield producing_task_key, filter_and_map(lambda t: t != "file"), filter_and_map(lambda t: t == "file")


def read_task_state(control_dir=None, no_state_file_ok=False):

    if control_dir is None:
        control_dir = os.environ["__control_dir"]

    glob_exp = os.path.join(control_dir, "state.*")
    state_file = list(glob.glob(glob_exp))

    if len(state_file) == 0:
        raise Exception(f"no state file in {control_dir}, {glob_exp}")
    if len(state_file) > 1:
        raise Exception(f"more than one state file found in {control_dir}, {glob_exp}")

    state_file = state_file[0]
    base_file_name = os.path.basename(state_file)
    name_parts = base_file_name.split(".")
    state_name = name_parts[1]

    if len(name_parts) == 3:
        step_number = int(name_parts[-1])
    else:
        step_number = 0

    return step_number, control_dir, state_file, state_name


def touch(fname):
    if os.path.exists(fname):
        os.utime(fname, None)
    else:
        open(fname, 'a').close()


def _append_to_history(control_dir, state_name, step_number=None):
    with open(os.path.join(control_dir, "history.tsv"), "a") as f:
        f.write(state_name)
        f.write("\t")
        f.write(datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%f'))
        f.write("\t")
        if step_number is None:
            f.write("")
        else:
            f.write(str(step_number))
        f.write("\n")


def _transition_state_file(state_file, next_state_name, step_number=None):

    logger.debug("_transition_state_file: %s", state_file)

    control_dir = os.path.dirname(state_file)

    if step_number is None:
        next_step_number = None
        next_state_basename = f"state.{next_state_name}"
    else:
        next_step_number = step_number
        next_state_basename = f"state.{next_state_name}.{next_step_number}"

    next_state_file = os.path.join(control_dir, next_state_basename)

    logger.info("will transition to: %s", next_state_basename)
    logger.debug("next_state_file: %s", next_state_file)

    os.rename(
        state_file,
        next_state_file
    )

    _append_to_history(control_dir, next_state_name, step_number)

    return next_state_file, next_step_number


def transition_to_step_started(state_file, step_number):
    return _transition_state_file(state_file, "step-started", step_number)


def transition_to_step_completed(state_file, step_number):
    state_file, step_number = _transition_state_file(state_file, "step-completed", step_number)
    return state_file, step_number + 1


def register_signal_handlers():

    def timeout_handler(s, frame):
        step_number, control_dir, state_file, state_name = read_task_state()
        _transition_state_file(state_file, "timed-out", step_number)
        _exit_process()

    logger.debug("will register signal handlers")

    signal.signal(signal.SIGINT, signal.SIG_IGN)
    signal.signal(signal.SIGHUP, signal.SIG_IGN)

    signal.signal(signal.SIGUSR1, timeout_handler)
    signal.signal(signal.SIGTERM, _terminate_descendants_and_exit)

    logger.debug("signal handlers registered")


def sign_files():
    file_list_to_sign = os.environ.get('__file_list_to_sign')

    if file_list_to_sign is None or file_list_to_sign == "":
        return

    sig_dir = os.path.join(os.environ['__control_dir'], 'out_sigs')

    Path(sig_dir).mkdir(exist_ok=True)

    def all_files():
        for f in file_list_to_sign.split(","):
            if os.path.exists(f):
                yield f

    def checksum_one_file(f):
        bf = os.path.basename(f)
        sig_file = os.path.join(sig_dir, f"{bf}.sig")
        cmd = f"sha1sum {f}"
        with PortablePopen(cmd.split(" ")) as p:
            p.wait_and_raise_if_non_zero()
            sha1sum = p.stdout_as_string()
            with open(sig_file, "w") as f:
                f.write(sha1sum)
            return True

    with ThreadPoolExecutor(max_workers=4) as executor:
        for b in executor.map(checksum_one_file, all_files()):
            assert b


def transition_to_completed(state_file):
    return _transition_state_file(state_file, "completed-unsigned")


def write_out_vars(out_vars):

    all_vars = [
        f"{k}={v}" for k, v in out_vars.items()
    ]

    with open(os.environ["__output_var_file"], "w") as f:
        f.write("\n".join(all_vars))

    logger.info("out vars written: %s", ",".join(all_vars))


def run_script(script, container=None):

    env = {**os.environ}

    dump_env = f'python3 -c "import os, json; print(json.dumps(dict(os.environ)))"'
    out = os.environ['__out_log']
    err = os.environ['__err_log']

    cmd = ["bash", "-c", f". {script} 1> {out} 2> {err} && {dump_env}"]

    if container is not None:
        container_full_path = os.path.join(os.environ['__containers_dir'], container)
        cmd = [
            "singularity",
            "exec",
            container_full_path,
        ] + cmd

        singularity_bindings = []

        root_dir_of_script = _root_dir(script)

        if _fs_type(root_dir_of_script) in ["autofs", "nfs", "zfs"]:
            singularity_bindings.append(f"{root_dir_of_script}:{root_dir_of_script}")

        if os.environ.get("__is_slurm"):
            scratch_dir = os.environ['SLURM_TMPDIR']
            env['__scratch_dir'] = scratch_dir
            root_of_scratch_dir = _root_dir(scratch_dir)
            singularity_bindings.append(f"{root_of_scratch_dir}:{root_of_scratch_dir}")

        if len(singularity_bindings) > 0:

            prev_singularity_bindings = env.get("SINGULARITY_BIND")

            if prev_singularity_bindings is not None and prev_singularity_bindings != "":
                bindings_prefix = f"{prev_singularity_bindings},"
            else:
                bindings_prefix = ""

            env["SINGULARITY_BIND"] = f"{bindings_prefix}{','.join(singularity_bindings)}"

    logger.info("run_script: %s", cmd)

    has_failed = False
    try:

        with PortablePopen(cmd, env=env) as p:
            p.wait_and_raise_if_non_zero()
            out = p.stdout_as_string()
            step_output_vars = json.loads(out)
            task_output_vars = dict(iterate_out_vars_from(os.environ["__output_var_file"]))

            for producing_task_key, var_metas, file_metas in parse_in_out_meta({
                k: v
                for k, v in os.environ.items()
                if k.startswith("__meta_")
            }):
                if producing_task_key != "":
                    continue

                for var_meta in var_metas:

                    name_in_producing_task, var_name, typez = var_meta
                    v = step_output_vars.get(var_name)

                    task_output_vars[var_name] = v
                    if v is not None:
                        os.environ[var_name] = v

            write_out_vars(task_output_vars)


    except Exception as ex:
        logger.exception(ex)
        has_failed = True
    finally:
        if has_failed:
            step_number, control_dir, state_file, state_name = read_task_state()
            _transition_state_file(state_file, "failed", step_number)
            _exit_process()


def launch_task_from_remote(task_key, is_slurm, wait_for_completion, drypipe_task_debug):

    pipeline_instance_dir = os.path.dirname(os.path.dirname(sys.argv[0]))

    control_dir = os.path.join(pipeline_instance_dir, '.drypipe', task_key)

    logger = create_task_logger(control_dir)

    out_sigs_dir = os.path.join(control_dir, 'out_sigs')
    touch(os.path.join(control_dir, 'output_vars'))

    work_dir = os.path.join(pipeline_instance_dir, 'output', task_key)
    scratch_dir = os.path.join(work_dir, "scratch")

    for d in [work_dir, out_sigs_dir, scratch_dir]:
        Path(d).mkdir(exist_ok=True, parents=True)

    for state_file in glob.glob(os.path.join(control_dir, "state.*")):
        os.remove(state_file)

    touch(os.path.join(control_dir, 'state.launched.0'))

    env = os.environ

    if is_slurm:
        cmd = os.path.join(control_dir, 'sbatch-launcher.sh')
        if wait_for_completion:
            env = {**env, "SBATCH_EXTRA_ARGS": "--wait", "DRYPIPE_TASK_DEBUG": str(drypipe_task_debug)}
            cmd = ["bash", "-c", cmd]
    else:
        if wait_for_completion:
            back_ground = ""
        else:
            back_ground = "&"
        scr = os.path.join(control_dir, 'task')
        cmd = ["nohup", "bash", "-c", f"python3 {scr} {back_ground}"]

    logger.info("launching from remote: %s", ' '.join(cmd))

    with open(os.path.join(control_dir, 'out.log'), 'w') as out:
        with open(os.path.join(control_dir, 'err.log'), 'w') as err:
            with PortablePopen(
                cmd,
                stdout=out,
                stderr=err,
                env=env
            ) as p:
                p.wait()
                if p.popen.returncode != 0:
                    logger.error(f"remote task launch failed: %s", sys.argv)
                print(str(p.popen.returncode))


def handle_main(task_func):
    is_kill_command = "kill" in sys.argv
    is_tail_command = "tail" in sys.argv
    wait_for_completion = "--wait" in sys.argv
    is_ps_command = "ps" in sys.argv

    is_run_command = not (is_tail_command or is_ps_command or is_kill_command)

    if is_run_command:
        _launch_task(task_func, wait_for_completion)
    else:

        def quit_command(p1, p2):
            logging.shutdown()
            exit(0)

        signal.signal(signal.SIGINT, quit_command)

        if is_tail_command:
            _tail_command()
        elif is_ps_command:
            _ps_command()
        elif is_kill_command:
            _kill_command()
        else:
            raise Exception(f"bad args: {sys.argv}")


def _load_pid_or_job_id(control_dir, pid_or_job_id):
    pid_file = os.path.join(control_dir, pid_or_job_id)
    if not os.path.exists(pid_file):
        return None
    else:
        with open(pid_file) as f:
            return int(f.read())


def load_pid(control_dir):
    return _load_pid_or_job_id(control_dir, "pid")


def load_slurm_job_id(control_dir):
    return _load_pid_or_job_id(control_dir, "slurm_job_id")


def print_pid_not_found(sloc):
    print(f"file {os.path.join(sloc, 'pid')} not found, task not running")


def _kill_command():
    sloc = os.environ["__script_location"]
    pid = load_pid(sloc)
    if pid is None:
        print_pid_not_found(sloc)
    else:
        os.kill(pid, signal.SIGTERM)


def _ps_command(pid=None):

    if pid is None:
        sloc = os.environ["__script_location"]
        pid = load_pid(sloc)
        if pid is None:
            print_pid_not_found(sloc)
            return

    res = list(ps_resources(pid))

    if len(res) == 0:
        return

    header, row = res

    sys.stdout.write("\t".join(header))
    sys.stdout.write("\n")
    sys.stdout.write("\t".join(str(v) for v in row))
    sys.stdout.write("\n")


def sizeof_fmt(num, suffix="B"):
    for unit in ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"]:
        if abs(num) < 1024.0:
            return f"{num:3.1f}{unit}{suffix}"
        num /= 1024.0
    return f"{num:.1f}Yi{suffix}"


def ps_resources(pid):

    ps_format = "size,%mem,%cpu,rss,rsz,etimes"

    def parse_ps_line(r):
        size, pmem, pcpu, rss, rsz, etimes = r.split()
        return int(size), float(pmem), float(pcpu), int(rss), int(rsz), int(etimes)

    with PortablePopen(["ps", f"--pid={pid}", f"--ppid={pid}", "o", ps_format, "--no-header"]) as p:
        p.wait()
        if p.popen.returncode != 0:
            return iter([])

        def f(t1, t2):
            size1, pmem1, pcpu1, rss1, rsz1, etimes1 = t1
            size2, pmem2, pcpu2, rss2, rsz2, etimes2 = t2
            return (
               size1 + size2,
               pmem1 + pmem2,
               pcpu1 + pcpu2,
               rss1 + rss2,
               rsz1 + rsz2,
               max(etimes1, etimes2)
            )

        yield ps_format.split(",")

        out = p.stdout_as_string().strip()
        if out != "":
            yield list(reduce(
                f,
                [parse_ps_line(line) for line in out.split("\n")],
                (0, 0, 0, 0, 0, 0)
            ))


def _tail_command():

    script_location = os.environ["__script_location"]

    all_logs = [
        os.path.join(script_location, log)
        for log in ["drypipe.log", "out.log", "err.log"]
    ]

    tail_cmd = ["tail", "-f"] + all_logs

    with PortablePopen(tail_cmd, stdout=sys.stdout, stderr=sys.stderr, stdin=sys.stdin) as p:
        p.wait()


def _launch_task(task_func, wait_for_completion):

    def task_func_wrapper():
        try:
            logger.debug("task func started")
            task_func()
            logger.info("task completed")
        except Exception as ex:
            logger.exception(ex)
        finally:
            _exit_process()

    if wait_for_completion:
        task_func_wrapper()
    else:
        is_slurm = os.environ.get("__is_slurm") == "True"
        if (not is_slurm) and os.fork() != 0:
            exit(0)
        else:
            sloc = os.environ['__script_location']
            if is_slurm:
                slurm_job_id = os.environ.get("SLURM_JOB_ID")
                logger.info("slurm job started, slurm_job_id=%s", slurm_job_id)
                with open(os.path.join(sloc, "slurm_job_id"), "w") as f:
                    f.write(str(os.getpid()))
            else:
                with open(os.path.join(sloc, "pid"), "w") as f:
                    f.write(str(os.getpid()))

            os.setpgrp()
            register_signal_handlers()
            Thread(target=task_func_wrapper).start()
            signal.pause()


def gen_input_var_exports(out=sys.stdout, env=os.environ):

    pipeline_instance_dir = env.get("__pipeline_instance_dir")

    task_key = env.get('__task_key')

    if pipeline_instance_dir is None:
        raise Exception("env variable __pipeline_instance_dir not set")

    if task_key is None:
        raise Exception("env variable __task_key not set")

    for producing_task_key, var_metas, file_metas in parse_in_out_meta({
        k: v
        for k, v in env.items()
        if k.startswith("__meta_")
    }):

        if producing_task_key == "":
            continue

        # control_dir = os.path.join(pipeline_instance_dir, ".drypipe", producing_task_key)

        # step_number, control_dir, state_file, state_name = read_task_state(control_dir, no_state_file_ok=True)

        # if not state_name == "completed":
        #    print(f"export __non_completed_dependent_task={producing_task_key}", file=out)
        #    return

        out_vars = dict(iterate_out_vars_from(
            os.path.join(pipeline_instance_dir, ".drypipe", producing_task_key, "output_vars")
        ))

        for name_in_producing_task, var_name, typez in var_metas:
            v = out_vars.get(name_in_producing_task)
            if v is not None:
                print(f"export {var_name}={v}", file=out)


def _root_dir(d):
    p = Path(d)
    return os.path.join(p.parts[0], p.parts[1])


def _fs_type(file):

    stat_cmd = f"stat -f -L -c %T {file}"
    with PortablePopen(stat_cmd.split()) as p:
        p.wait_and_raise_if_non_zero()
        return p.stdout_as_string().strip()


def set_singularity_bindings():

    def root_dir(d):
        p = Path(d)
        return os.path.join(p.parts[0], p.parts[1])

    pipeline_code_dir = os.environ['__pipeline_code_dir']
    root_of_pipeline_code_dir = root_dir(pipeline_code_dir)
    stat_cmd = f"stat -f -L -c %T {root_of_pipeline_code_dir}"
    with PortablePopen(stat_cmd.split()) as p:
        p.wait_and_raise_if_non_zero()
        fs_type = p.stdout_as_string()

        bind_list = []

        if fs_type in ["autofs", "nfs", "zfs"]:
            bind_list.append(f"{root_of_pipeline_code_dir}:{root_of_pipeline_code_dir}")

        if os.environ.get("__is_slurm"):
            scratch_dir = os.environ.get("__scratch_dir")
            root_of_scratch_dir = root_dir(scratch_dir)
            bind_list.append(f"{root_of_scratch_dir}:{root_of_scratch_dir}")

        prev_singularity_bind = os.environ.get("SINGULARITY_BIND")
        if prev_singularity_bind is not None:
            bind_list.append(prev_singularity_bind)

        os.environ['SINGULARITY_BIND'] = ",".join(bind_list)


def terminate_all():
    scancel_all()
    segterm_all()


def scancel_all():
    pass


def segterm_all():
    pass


def handle_script_lib_main():

    if "gen-input-var-exports" in sys.argv:
        gen_input_var_exports()
    elif "launch-task-from-remote" in sys.argv:
        task_key = sys.argv[2]
        is_slurm = "--is-slurm" in sys.argv
        wait_for_completion = "--wait-for-completion" in sys.argv
        drypipe_task_debug = "--drypipe-task-debug" in sys.argv
        launch_task_from_remote(task_key, is_slurm, wait_for_completion, drypipe_task_debug)
    elif "terminate-all" in sys.argv:
        terminate_all()
    elif "scancel-all" in sys.argv:
        terminate_all()
    elif "segterm-all" in sys.argv:
        segterm_all()
    else:
        raise Exception('invalid args')



import glob
import os
import time
import pathlib
import mmap
import copy
from datetime import datetime
from itertools import groupby

from dry_pipe.actions import TaskAction
from dry_pipe.script_lib import parse_in_out_meta, PortablePopen, load_pid, load_slurm_job_id

"""

 /x/y/z/.drypipe/<task_key>/state.<state_name>.<next_or_current_step>
 
 All state files of a pipeline can be fetched with f"{pipeline_instance_dir}.drypipe/*/state.*
 
 And the name of files can 
"""

VALID_TRANSITIONS_BY_JANITORS = {

    "":                    ["waiting-for-deps"],
    "waiting-for-deps":    ["prepared"],
    "prepared":            ["queued",             "queued-for-upload"],

    "queued-for-upload":   ["upload-started"],
    "upload-started":      ["upload-completed",   "upload-failed"],
    "upload-completed":    ["queued"],

    "queued-for-download": ["download-started"],
    "download-started":    ["download-completed", "download-failed"],
    "download-completed":  ["completed"],

    "upload-failed":       ["upload-started"],
    "download-failed":     ["download-started"],

    "completed-unsigned":  ["completed", "queued-for-download"]
}

VALID_TRANSITIONS_BY_TASK_WORKERS = {
    "queued":              ["launched",           "scheduled"],
    "launched":            ["step-started",       "crashed", "queued-for-download"],
    "scheduled":           ["step-started",       "crashed"],
    "step-started":        ["step-completed",     "crashed",            "failed",             "timed-out"],
    "step-completed":      ["completed-unsigned"]
}

VALID_TRANSITIONS_BY_TASK_ADMIN = {
    "crashed":             ["launched",           "scheduled",          "step-started"],
    "failed":              ["queued",             "ignored"],
    "timed-out":           ["queued"],
    "killed":              ["queued"],
    "completed":           ["waiting-for-deps"],
    "ignored":             ["waiting-for-deps"]
}

VALID_TRANSITIONS = {
    ** VALID_TRANSITIONS_BY_JANITORS,
    ** VALID_TRANSITIONS_BY_TASK_WORKERS,
    ** VALID_TRANSITIONS_BY_TASK_ADMIN
}

STATE_EQUIVALENCE_FOR_DISPLAY = {

    "waiting-for-deps":    "waiting",
    "prepared":            "waiting",
    "upload-failed":       "waiting",
    "download-failed":     "waiting",

    "launched":            "scheduled",
    "scheduled":           "scheduled",

    "download-started":    "transit",
    "upload-started":      "transit",
    "queued-for-upload":   "transit",
    "upload-completed":    "transit",
    "queued-for-download": "transit",
    "download-completed":  "transit",
    "queued":              "transit",
    "step-completed":      "transit",

    "completed-unsigned": "running",
    "step-started": "running",

    "completed":           "completed",

    "failed":              "failed",
    "timed-out":           "timed-out",
    "killed":              "killed",
#    "ignored":             "ignored",
#    "crashed":             "crashed"
}

STATE_EQUIVALENCE_FOR_DISPLAY_WEB = {

    "waiting-for-deps":    "waiting",
    "prepared":            "waiting",
    "upload-failed":       "waiting",
    "download-failed":     "waiting",
    "scheduled":           "waiting",
    "killed":              "waiting",

    "launched":            "running",
    "completed-unsigned":  "running",
    "step-started":        "running",
    "download-started":    "running",
    "upload-started":      "running",
    "queued-for-upload":   "running",
    "upload-completed":    "running",
    "queued-for-download": "running",
    "download-completed":  "running",
    "queued":              "running",
    "step-completed":      "running",

    "completed":           "completed",

    "failed":              "failed",
    "timed-out":           "failed",
    "crashed":             "failed",
    "ignored":             "ignored"
}


NON_TERMINAL_STATES = [

    "prepared",
    "launched",
    "scheduled",
    "completed-unsigned",
    "step-started",
    "download-started",
    "upload-started",
    "queued-for-upload",
    "upload-completed",
    "queued-for-download",
    "download-completed",
    "queued",
    "step-completed"
]


class TaskState:

    @staticmethod
    def display_state_groups():

        return [
            (group_display_label, [t[0] for t in state_names_group_iterator])
            for group_display_label, state_names_group_iterator
            in groupby(STATE_EQUIVALENCE_FOR_DISPLAY.items(), key=lambda t: t[1])
        ]

    @staticmethod
    def state_display_names():
        return [
            display_name
            for display_name, state_names
            in TaskState.display_state_groups()
        ]

    state_display_names_web = list(set(STATE_EQUIVALENCE_FOR_DISPLAY_WEB.values()))

    @staticmethod
    def display_groups_of_state_names():
        return [
            list(state_names)
            for display_name, state_names
            in TaskState.display_state_groups()
        ]

    @staticmethod
    def fetch(pipeline, pattern, pipeline_instance_dir=None):

        if pipeline_instance_dir is None:
            pipeline_instance_dir = pipeline.pipeline_instance_dir

        for f in glob.glob(os.path.join(pipeline_instance_dir, f".drypipe/*/{pattern}")):
            yield TaskState(f)

    @staticmethod
    def from_task_control_dir(pipeline_instance_dir, task_key):

        state_file_pattern = os.path.join(pipeline_instance_dir, ".drypipe", task_key, "state.*")

        g = glob.glob(state_file_pattern)

        if len(g) == 0:
            return None

        return TaskState(g[0])

    @staticmethod
    def fetch_all(pipeline_instance_dir):
        return TaskState.fetch(None, "state.*", pipeline_instance_dir)

    @staticmethod
    def failed_task_states(pipeline_instance_dir):
        return TaskState.fetch(None, "state.failed.*", pipeline_instance_dir)

    @staticmethod
    def queued_for_upload_task_states(pipeline):
        return TaskState.fetch(pipeline, "state.queued-for-upload.*")

    @staticmethod
    def queued_for_dowload_task_states(pipeline):
        return TaskState.fetch(pipeline, "state.queued-for-download.*")

    @staticmethod
    def queued_for_remote_execution_task_states(pipeline):
        return TaskState.fetch(pipeline, "state.upload-completed.*")

    @staticmethod
    def queued_task_states(pipeline):
        return TaskState.fetch(pipeline, "state.queued.*")

    @staticmethod
    def prepared_task_states(pipeline):
        return TaskState.fetch(pipeline, "state.prepared.*")

    @staticmethod
    def waiting_for_deps_task_states(pipeline):
        return TaskState.fetch(pipeline, "state.waiting-for-deps")

    @staticmethod
    def count_running_local(pipeline):
        return sum([
            1 for f in glob.glob(os.path.join(pipeline.pipeline_instance_dir, f".drypipe/*/state.step-started*"))
        ])


    @staticmethod
    def completed_stale_task_states(pipeline):

        for f in glob.glob(os.path.join(pipeline.pipeline_instance_dir, f".drypipe/*/in.sig.stale")):
            task_control_dir = os.path.dirname(f)
            state_file = os.path.join(task_control_dir, "state.completed")

            if not os.path.exists(state_file):
                raise Exception(f"file {f} should not exist given that task is not in state.completed")

            yield TaskState(state_file)

    @staticmethod
    def completed_unsigned_task_states(pipeline):
        return TaskState.fetch(pipeline, "state.completed-unsigned")

    @staticmethod
    def reset_stalled_transfers(pipeline):

        for task_state in TaskState.fetch(pipeline, "state.upload-started"):
            task_state.reset_upload()

        for task_state in TaskState.fetch(pipeline, "state.download-started"):
            task_state.reset_download()

    @staticmethod
    def create_non_existing(task_control_dir):

        p = pathlib.Path(os.path.join(task_control_dir, "state.waiting-for-deps"))

        p.touch(exist_ok=False)

        return TaskState(p.absolute())

    @staticmethod
    def parse_history_timestamp(s):
        return datetime.strptime(s, '%Y-%m-%dT%H:%M:%S.%f')

    def __init__(self, state_file_name_with_abs_path, step_number=None):
        self._state_file_name_with_abs_path = state_file_name_with_abs_path
        self._control_dir = os.path.dirname(state_file_name_with_abs_path)
        self.task_key = os.path.split(self._control_dir)[-1]
        self._base_file_name = os.path.basename(state_file_name_with_abs_path)

        name_parts = self._base_file_name.split(".")

        self.state_name = name_parts[1]

        if VALID_TRANSITIONS.get(self.state_name) is None:
            raise Exception(f"file {state_file_name_with_abs_path} has invalid state: '{self.state_name}'")
        elif self.is_waiting_for_deps() or self.is_completed() or self.is_completed_unsigned():
            self._step_number = 0
        elif len(name_parts) == 3:
            self._step_number = int(name_parts[-1])
        else:
            self._step_number = step_number

    def __repr__(self):
        return self.base_file_name()

    def pid(self):
        return load_pid(self.control_dir())

    def slurm_job_id(self):
        return load_slurm_job_id(self.control_dir())

    def gen_meta_dict(self):
        task_env = os.path.join(self.control_dir(), "task-env.sh")

        def gen():
            with open(task_env) as f:
                for line in f.readlines():
                    if line.startswith("export __meta_"):
                        yield tuple(line[7:].split("="))
                    elif line.startswith("export END_META"):
                        break

        return dict(gen())

    def is_all_deps_ready(self):

        pid = self.pipeline_instance_dir()
        for task_key, _, _ in parse_in_out_meta(self.gen_meta_dict()):
            if task_key == "":
                continue
            task_state = TaskState.from_task_control_dir(pid, task_key)
            if not task_state.is_completed():
                return False

        return True

    def pipeline_instance_dir(self):
        return os.path.dirname(os.path.dirname(self.control_dir()))

    def base_file_name(self):
        return self._base_file_name

    def abs_file_name(self):
        return self._state_file_name_with_abs_path

    def control_dir(self):
        return self._control_dir

    def is_prepared(self):
        return self.state_name == "prepared"

    def is_queued(self):
        return self.state_name == "queued"

    def is_launched(self):
        return self.state_name == "launched"

    def is_scheduled(self):
        return self.state_name == "scheduled"

    def is_completed(self):
        return self.state_name == "completed"

    def is_failed(self):
        return self.state_name == "failed"

    def is_timed_out(self):
        return self.state_name == "timed-out"

    def is_waiting_for_deps(self):
        return self.state_name == "waiting-for-deps"

    def is_step_started(self):
        return self.state_name == "step-started"

    def is_step_completed(self):
        return self.state_name == "step-completed"

    def is_completed_unsigned(self):
        return self.state_name == "completed-unsigned"

    def is_killed(self):
        return self.state_name == "killed"

    def is_input_signature_changed(self):
        return os.path.exists(os.path.join(self._control_dir, "in.sig.changed"))

    def step_number(self):
        return self._step_number

    def tail_err_if_failed(self, n):

        if self.is_failed():
            err_file = os.path.join(self._control_dir, "err.log")
            return tail(err_file, n)
        else:
            return None

    def tail_out_err_drypipe(self):
        tails = [
            tail(os.path.join(self.control_dir(), d), 20)
            for d in ["out.log", "err.log", "drypipe.log"]
        ]

        return tuple(tails)


    def _ensure_transition_valid(self, next_state):
        valid_next_states = VALID_TRANSITIONS.get(self.state_name)

        if valid_next_states is None:
            raise Exception(f"task {self.task_key} is in unknown state {self.state_name}")

        if next_state not in valid_next_states:
            raise Exception(f"task {self.task_key} can't transition from '{self.state_name}' to '{next_state}'")

    def _transition(self, next_state, suffix="", col_4="", force=False):

        if not force:
            self._ensure_transition_valid(next_state)

        last_update_time = datetime.fromtimestamp(os.stat(self._state_file_name_with_abs_path).st_ctime)

        state_to_dump = self.state_name

        if self.state_name == "completed-unsigned":
            state_to_dump = "completed"

        history_line = [
            state_to_dump,
            last_update_time.strftime('%Y-%m-%dT%H:%M:%S.%f'),
            "", #str(getattr(self, "_step_number", "")),
            col_4
        ]

        with open(os.path.join(self._control_dir, "history.tsv"), "a") as f:
            f.write("\t".join(history_line)+"\n")

        next_file = os.path.join(self._control_dir, f"state.{next_state}{suffix}")
        os.rename(
            os.path.join(self._control_dir, self._base_file_name),
            next_file
        )

        return next_file

    def _step_and_retry_suffix(self, step=None):
        s = step if step is not None else self._step_number
        return f".{s}"

    def reset_upload(self):
        self._transition("queued-for-upload", ".0.0.3", force=True)

    def reset_download(self):
        self._transition("queued-for-download", ".0.0.3", force=True)

    def transition_to_waiting_for_deps(self):
        self._transition("waiting-for-deps")

    def transition_to_prepared(self, force=False):
        #TODO: eliminate prepared state
        self._transition("prepared", f".{self.step_number()}", force=force)

        from dry_pipe import Task
        task = Task.load_from_task_state(self)
        task_state = task.get_state()
        if task.is_remote():
            task_state.transition_to_queued_remote_upload()
        else:
            task_state.transition_to_queued()

    def transition_to_queued_remote_upload(self):
        self._transition("queued-for-upload", ".0.0.3")

    def transition_to_queued_remote_download(self, pipeline_instance_dir, task_key):

        state_file_expr = os.path.join(pipeline_instance_dir, ".drypipe", task_key, "state.*")

        for f in glob.glob(state_file_expr):
            task_state = TaskState(f)

            task_state._transition("queued-for-download", ".0.0.3", force=True)
            return

        raise Exception(f"expected state file {state_file_expr} to exist")

    def assign_remote_state_to_local_state_file(self, pipeline_instance_dir, remote_task_state):

        control_dir = os.path.join(pipeline_instance_dir, ".drypipe", remote_task_state.task_key)

        state_file_expr = os.path.join(control_dir, "state.*")

        for local_state_file in glob.glob(state_file_expr):

            if local_state_file != os.path.join(control_dir, remote_task_state.base_file_name()):
                os.rename(
                    local_state_file,
                    os.path.join(control_dir, remote_task_state.base_file_name())
                )
            return

        raise Exception(f"expected state file {state_file_expr} to exist")


    def transition_to_queued(self, increment_retry=True, step=None, force=False):

        if self.is_failed():
            c = copy.copy(self)
            c._transition("queued", c._step_and_retry_suffix(step), force=force)
        else:
            self._transition("queued", self._step_and_retry_suffix(step), force=force)

    def transition_to_killed(self):
        self._transition("killed", self._step_and_retry_suffix(), force=True)

    def transition_to_launched(self, executer, task, wait_for_completion=False, fail_silently=False):
        self._transition("launched", self._step_and_retry_suffix())
        task.launch(executer, wait_for_completion, fail_silently=fail_silently)

    def transition_to_completed(self, task):
        task.verify_output_files_produced()
        task.write_output_signature_file()
        self._transition("completed")

    def transition_to_upload_completed(self):
        return TaskState(self._transition("upload-completed"), 0)

    def transition_to_upload_started(self):
        return TaskState(self._transition("upload-started"))

    def transition_to_download_completed(self):
        return TaskState(self._transition("download-completed"), 0)

    def transition_to_download_started(self):
        return TaskState(self._transition("download-started"))

    def action_if_exists(self):
        return TaskAction.load_from_task_control_dir(self._control_dir)

    def history_steps_time_intervals(self):

        rows = [
            (int(r[2]), r[0], TaskState.parse_history_timestamp(r[1]))
            for r in self.load_history_rows()
            if r[0] in ["step-started", "step-completed"]
        ]

        for step, rows in groupby(rows, key=lambda t: t[0]):
            rows = list(rows)
            step_started = [s[2] for s in rows if s[1] == "step-started"]
            step_started = None if len(step_started) == 0 else step_started[0]
            step_completed = [s[2] for s in rows if s[1] == "step-completed"]
            step_completed = None if len(step_completed) == 0 else step_completed[0]
            yield rows[0][0], step_started, step_completed


    def load_history_rows(self, history_tsv_as_string=None):

        if history_tsv_as_string is None:
            h = os.path.join(self._control_dir, "history.tsv")
            if os.path.exists(h):
                with open(h) as f:
                    history_tsv_as_string = f.read()
            else:
                history_tsv_as_string = ""

        for line in history_tsv_as_string.split("\n"):
            if line != "":
                yield line.split("\t")

    def validate_history(self):

        rows = list(self.load_history_rows())

        for i in range(0, len(rows) - 1):
            state1 = rows[i][0]
            state2 = rows[i + 1][0]
            if state1 not in VALID_TRANSITIONS:
                yield f"invalid state: {state1}"
            else:
                possible_next_states = VALID_TRANSITIONS[state1]
                if state2 not in possible_next_states:
                    yield f"{state1} can't be followed by {state2}"

    def as_json(self):

        def file_content(file):
            f = os.path.join(self.control_dir(), file)
            if not os.path.exists(f):
                return None

            return f"tail -1000\n {tail(f, lines=1000)}"

        action = TaskAction.load_from_task_state(self)

        action = None if action is None else action.action_name

        snapshot_time = int(time.time_ns())

        return {
            'key': self.task_key,
            'state': self.state_name,
            'step': self.step_number(),
            'out': file_content("out.log"),
            'err': file_content("err.log"),
            'control_err': file_content("control-err.log"),
            'history': list(self.load_history_rows()),
            'action': action,
            'snapshot_time': snapshot_time
        }


def tail(filename, lines=20, line_limit=1000):

    if not os.path.exists(filename):
        return ""

    cmd = f"tail -{lines} {filename}"

    with PortablePopen(cmd.split(" ")) as p:
        p.wait_and_raise_if_non_zero()

        def read():
            for line in p.read_stdout_lines():
                if len(line) <= line_limit:
                    yield line
                else:
                    yield f"{line[0:line_limit]}...(line was truncated as it's length exceeded {line_limit} chars)\n"

        return "".join(read())



import asyncio
import logging
import os
import pathlib
import time
import weakref
from datetime import datetime
from threading import Thread

import socketio
import json
import uvicorn

from dry_pipe import Task
from dry_pipe.actions import TaskAction
from dry_pipe.janitors import DaemonThreadHelper
from dry_pipe.monitoring import PipelineMetricsTable, fetch_task_group_metrics
from dry_pipe.pipeline_state import PipelineState
from dry_pipe.task_state import TaskState

logger = logging.getLogger(__name__)

DRYPIPE_WEBSOCKET_NAMESPACE = "/drypipe"


def task_room_name(pipeline_dir, task_key):
    return f"task,{pipeline_dir},{task_key}"


class PipelineUIServer:

    instance = None

    @staticmethod
    def init_multi_pipeline_instance(instances_dir):
        PipelineUIServer.instance = PipelineUIServer(instances_dir=instances_dir)
        return PipelineUIServer.instance

    @staticmethod
    def init_single_pipeline_instance(single_pipeline_state):
        PipelineUIServer.instance = PipelineUIServer(single_pipeline_state=single_pipeline_state)
        return PipelineUIServer.instance

    def is_single_pipeline_mode(self):
        return self.single_pipeline_state is not None

    def __init__(self, instances_dir=None, single_pipeline_state=None):
        self.instances_dir = instances_dir
        self.single_pipeline_state = single_pipeline_state

        if single_pipeline_state is not None:
            self.pipeline_states = [single_pipeline_state]
        else:
            self.pipeline_states = [
                pipeline_state
                for pipeline_state in PipelineState.iterate_from_instances_dir(instances_dir)
            ]

        self.shutdown = False

        def call_async(f):
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

            async def asyncify():
                await f()

            loop.run_until_complete(asyncify())
            loop.close()

        def work():

            daemon_thread_helper = DaemonThreadHelper(
                logging.getLogger(f"{__name__}.ui_server"),
                min_sleep=2, max_sleep=6, pipelines=None
            )

            while not self.shutdown:

                daemon_thread_helper.begin_round()

                try:
                    rooms = sio.manager.rooms

                    if DRYPIPE_WEBSOCKET_NAMESPACE in rooms:
                        for pipeline_state in self.iterate_pipeline_states():

                            #pipeline_holder = self.get_and_maybe_load_pipeline_holder(pipeline_state)

                            def is_room_active(room_name):
                                return DRYPIPE_WEBSOCKET_NAMESPACE in rooms and \
                                       room_name in rooms[DRYPIPE_WEBSOCKET_NAMESPACE]

                            if is_room_active(f"counts-{pipeline_state.dir_basename()}"):

                                call_async(
                                    self.prepare_emit_counts_message(pipeline_state.instance_dir())
                                )

                                daemon_thread_helper.register_work()
                                #pipeline_holder.inc_work_counter()

                    if DRYPIPE_WEBSOCKET_NAMESPACE in rooms:
                        for room_name in list(rooms[DRYPIPE_WEBSOCKET_NAMESPACE]):
                            if room_name is not None and room_name.startswith("task,"):

                                t, instance_dir, task_key = room_name.split(",")

                                call_async(
                                    self.prepare_emit_task_details_message(
                                        os.path.join(self.instances_dir, instance_dir),
                                        task_key,
                                        room_name
                                    )
                                )

                                daemon_thread_helper.register_work()

                    daemon_thread_helper.end_round()
                except Exception as ex:
                    daemon_thread_helper.handle_exception_in_daemon_loop(ex)

        self.thread = Thread(target=work)

    def iterate_pipeline_states(self):
        if self.is_single_pipeline_mode:
            for pipeline_state in self.pipeline_states:
                yield pipeline_state
        else:
            for s in PipelineState.iterate_from_instances_dir(self.instances_dir):
                yield s

    def full_pipeline_instance_dir(self, pipeline_dir):
        if self.is_single_pipeline_mode():
            return self.single_pipeline_state.instance_dir()
        else:
            return os.path.join(self.instances_dir, pipeline_dir)

    def prepare_emit_counts_message(self, pipeline_instance_dir):

        pipeline_state = PipelineState.from_pipeline_instance_dir(pipeline_instance_dir)
        instance_dir = pipeline_state.dir_basename()

        actions_by_task_key = {
            task_action.task_key: task_action
            for task_action in TaskAction.fetch_all_actions(pipeline_state.instance_dir())
        }

        def create_visitor():

            partial_task_infos = []

            def visit_task(task_state):

                o = {
                    "key": task_state.task_key,
                    "state_name": task_state.state_name,
                    "step": task_state.step_number()
                }

                err_tail = task_state.tail_err_if_failed(3)

                if err_tail is not None:
                    o["err_tail"] = err_tail

                action = actions_by_task_key.get(task_state.task_key)

                if action is not None:
                    o["action"] = action.action_name

                partial_task_infos.append(o)

            return partial_task_infos, visit_task

        partial_task_infos, task_state_visitor = create_visitor()

        snapshot_time = int(time.time_ns())

        data = {
            # SHOULD SEND DAG HERE ?
            # "dag": pipeline.summarized_dependency_graph(),
            "tsv": all_task_states_as_tsv(pipeline_state.instance_dir(), task_state_visitor),
            "pipelineDir": pipeline_state.dir_basename(),
            "partial_task_infos": partial_task_infos,
            "snapshot_time": snapshot_time
        }

        return lambda: sio.emit(
            'latestPipelineDetailedState',
            data,
            room=f"counts-{instance_dir}",
            namespace=DRYPIPE_WEBSOCKET_NAMESPACE
        )

    def prepare_emit_task_details_message(self, pipeline_instance_dir, task_key, room_name):

        task_state = TaskState.from_task_control_dir(pipeline_instance_dir, task_key)

        data = task_state.as_json()

        return lambda: sio.emit(
            'latestTaskDetails',
            data,
            room=room_name,
            namespace=DRYPIPE_WEBSOCKET_NAMESPACE
        )

    def start(self):
        self.thread.start()

    def stop(self):
        self.shutdown = False
        self.thread.join()

    def all_pipeline_states(self):
        return [
            {
                "dir": pipeline_state.dir_basename(),
                "state": pipeline_state.state_name,
                "totals": pipeline_state.totals_row()
            }
            for pipeline_state in self.pipeline_states
        ]


def all_task_states_as_tsv(pipeline_instance_dir, task_state_visitor=None):

    header, table_body, footer = PipelineMetricsTable.detailed_table_from_task_group_metrics(
        fetch_task_group_metrics(
            pipeline_instance_dir,
            Task.key_grouper,
            task_state_visitor=task_state_visitor
        )
    )

    def format_cell(c):
        return str(c or 0)

    res = "\n".join([
        "\t".join([
            format_cell(cell) for cell in row
        ])
        for row in table_body
    ])

    header = '\t'.join(header[1:-1])

    return f"{header}\n{res}"


def json_response(func):

    def enc(s):
        return s.encode("utf-8")

    body = enc(json.dumps(func()))

    headers = {
        'type': 'http.response.start',
        'status': 200,
        'headers': [
            [b'Content-Type', b'application/json'],
            [b'Content-Length', enc(str(len(body)))]
        ],
    }

    def body_func():
        return {
            'type': 'http.response.body',
            'body': body
        }

    return headers, body_func


def text_response(func):

    headers = {
        'type': 'http.response.start',
        'status': 200,
        'headers': [
            [b'content-type', b'text/plain'],
        ],
    }

    def body_func():
        return {
            'type': 'http.response.body',
            'body': func().encode("utf-8")
        }

    return headers, body_func


def map_static_file(static_file):

    the_static_file = os.path.join(
        os.path.abspath(os.path.dirname(__file__)),
        "ui",
        static_file
    )

    if not os.path.exists(the_static_file):
        the_static_file = os.path.join(
            os.path.abspath(os.path.dirname(os.path.dirname(__file__))),
            "ui",
            "dist",
            static_file
        )

    return the_static_file


async def serve_static_file(static_file, ct, send, receive):

    the_static_file = map_static_file(static_file)

    event = await receive()

    with open(the_static_file, 'rb') as f:
        payload = f.read()

    await send({'type': 'http.response.start',
                'status': 200,
                'headers': [(b'Content-Type', ct.encode('utf-8'))]})

    await send({'type': 'http.response.body',
                'body': payload})


class App:

    async def __call__(self, scope, receive, send):

        if scope["type"] == "lifespan":
            return

        assert scope['type'] == 'http'

        path = scope["path"]

        if path == "/":
            await serve_static_file("index.html", 'text/html', send, receive)
            return

        response = None

        if scope["method"] == "GET":

            if path == "/drypipe/pipelines":
                response = json_response(lambda: PipelineUIServer.instance.all_pipeline_states_as_json())

            elif path.startswith("/drypipe/pipeline/"):

                if "/action/" in path:

                    response = json_response(lambda: PipelineUIServer.instance.all_pipeline_states_as_json())

        if response is None:
            response = text_response(lambda: "Bad Route")

        headers, body = response

        await send(headers)

        b = body()

        await send(b)


sio = socketio.AsyncServer(async_mode='asgi')


@sio.on('join', namespace=DRYPIPE_WEBSOCKET_NAMESPACE)
async def join(sid, message):
    room = message['room']
    #print(f"entering {room}")
    sio.enter_room(sid, room, namespace=DRYPIPE_WEBSOCKET_NAMESPACE)

    if PipelineUIServer.instance is not None:

        pipeline_ui_server = PipelineUIServer.instance

        if room.startswith("counts-"):
            instance_dir = room[7:]
            #pipeline_holder = pipeline_ui_server.get_and_maybe_load_pipeline_holder(instance_dir=instance_dir)

            pid = os.path.join(pipeline_ui_server.instances_dir, instance_dir)
            emit_call = pipeline_ui_server.prepare_emit_counts_message(pid)
            await emit_call()

        if room.startswith("task,"):
            t, instance_dir, task_key = room.split(",")

            logger.debug("new observer for task %s in pipeline %s", task_key, instance_dir)

            pid = os.path.join(pipeline_ui_server.instances_dir, instance_dir)

            #pipeline_holder = pipeline_ui_server.get_and_maybe_load_pipeline_holder(instance_dir=instance_dir)
            emit_call = pipeline_ui_server.prepare_emit_task_details_message(pid, task_key, room)
            await emit_call()


@sio.on('leave', namespace=DRYPIPE_WEBSOCKET_NAMESPACE)
async def leave(sid, message):
    room = message['room']

    logger.debug("stop observing %s", room)

    sio.leave_room(sid, room, namespace=DRYPIPE_WEBSOCKET_NAMESPACE)


@sio.on('submitTaskAction', namespace=DRYPIPE_WEBSOCKET_NAMESPACE)
async def task_action_submitted(sid, message):

    task_key = message['task_key']
    pipeline_dir = message['pipeline_dir']
    action_name = message['action_name']
    is_cancel = message.get("is_cancel")

    if PipelineUIServer.instance is not None:

        pipeline_ui_server = PipelineUIServer.instance

        pipeline_instance_dir = pipeline_ui_server.full_pipeline_instance_dir(pipeline_dir)

        try:

            if is_cancel is not None and is_cancel:
                action = TaskAction.load_from_task_control_dir(
                    os.path.join(pipeline_instance_dir, ".drypipe", task_key)
                )

                if action is not None:
                    action.delete()

                await sio.emit(
                    'taskActionSubmitted',
                    message,
                    room=task_room_name(pipeline_dir, task_key),
                    namespace=DRYPIPE_WEBSOCKET_NAMESPACE
                )

                emit_call = pipeline_ui_server.prepare_emit_task_details_message(
                    pipeline_instance_dir, task_key, task_room_name(pipeline_dir, task_key)
                )

                await emit_call()

            else:
                TaskAction.submit(pipeline_instance_dir, task_key, action_name, message.get("step"))
                await sio.emit(
                    'taskActionSubmitted',
                    message,
                    room=task_room_name(pipeline_dir, task_key),
                    namespace=DRYPIPE_WEBSOCKET_NAMESPACE
                )
        except Exception as ex:
            await sio.emit(
                "serverSideException",
                {"message": f"{ex}"},
                room=sid,
                namespace=DRYPIPE_WEBSOCKET_NAMESPACE
            )

    #print(f"leaving {room}")

@sio.on('close room')
async def close(sid, message):
    await sio.close_room(message['room'])


@sio.on('disconnect request')
async def disconnect_request(sid):
    await sio.disconnect(sid)


app = socketio.ASGIApp(
    sio,
    other_asgi_app=App(),
    static_files={
        "/main.js": map_static_file("dist/main.js"),
        "/main.js.map": map_static_file("dist/main.js.map")
    }
)


def is_port_in_use(bind_address, port):
    import socket
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex((bind_address, port)) == 0


def serve(bind_address, port):

    if is_port_in_use(bind_address, port):
        print(f" port {port} is used by another application, use --port=X to listen on another port")
        return

    print(f"Web Monitor running on: http://{bind_address}:{port} (Press CTRL+C to quit)")

    uvicorn.run(
        #"dry_pipe.server:app",
        app,
        host=bind_address, port=port,
        log_level="error"
        #reload=True,
        #reload_dirs=["/home/maxou/dev/dry_pipe/dry_pipe"],
        #reload_delay=2.0
    )



import os
import tempfile
import logging.config

from dry_pipe import Executor, bash_shebang
from dry_pipe.internals import flatten
from dry_pipe.script_lib import PortablePopen, create_task_logger
from dry_pipe.task_state import TaskState
from dry_pipe.utils import perf_logger_timer

logger_ssh = logging.getLogger(f"{__name__}.ssh")

class ScpUploadException(Exception):
    pass


class RemoteSSH(Executor):

    def __init__(
        self, ssh_username, ssh_host, remote_base_dir, key_filename, before_execute_bash
    ):
        if key_filename is None:
            raise Exception("key_filename can't be none")

        self.before_execute_bash = before_execute_bash
        self.remote_base_dir = remote_base_dir
        self.ssh_host = ssh_host
        self.ssh_username = ssh_username
        self.key_filename = key_filename
        self.rsync_containers = True
        self.slurm = None
        self.ssh_timeout_in_seconds = 60
        self.ssh_client = None

    def is_remote(self):
        return True

    def user_at_host(self):
        return f"{self.ssh_username}@{self.ssh_host}"

    def server_connection_key(self):
        return f"{self.ssh_username}-{self.ssh_host}"

    def connect(self):
        from pssh.clients import SSHClient
        pkey = os.path.expanduser(self.key_filename) \
            if self.key_filename is not None and self.key_filename.startswith("~") else self.key_filename
        try:
            with perf_logger_timer("RemoteSSH.connect") as t:
                self.ssh_client = SSHClient(
                    self.ssh_host,
                    self.ssh_username,
                    pkey=pkey
                )
            logger_ssh.info("new ssh connection established %s at %s", self.ssh_username, self.ssh_host)
        except Exception as ex:
            logger_ssh.exception("%s", [self.ssh_host, self.ssh_username, pkey])
            raise ex

    def invoke_remote(self, cmd, bash_error_ok=False):
        logger_ssh.debug("will invoke '%s' at %s", cmd, self.ssh_host)

        with perf_logger_timer("RemoteSSH.invoke_remote", cmd) as t:

            host_out = self.ssh_client.run_command(cmd)
            stdout_txt = "\n".join(list(host_out.stdout))
            stderr_lines = list(host_out.stderr)

        if not bash_error_ok and host_out.exit_code != 0:
            stderr_txt = "\n".join(stderr_lines)
            raise Exception(f"remote call failed {host_out.exit_code}, '{cmd}'\n{stderr_txt}\non {self}")

        logger_ssh.debug(f"invocation of '%s' at {self} returned '%s'", cmd, stdout_txt)

        return stdout_txt

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def close(self):
        self.ssh_client.disconnect()

    @staticmethod
    def exception_is_recoverable_with_connection_reset(ex):
        from pssh.exceptions import SSHException, SSHError, SFTPError
        if isinstance(ex, SSHException) or \
           isinstance(ex, SFTPError) or \
           isinstance(ex, SSHError):
            return True
        else:
            return False

    def fetch_remote_task_states(self, pipeline):
        with perf_logger_timer("RemoteSSH.fetch_remote_task_states") as t:
            remote_pid_basename = os.path.basename(pipeline.pipeline_instance_dir)

            cmd = f"find {self.remote_base_dir}/{remote_pid_basename}/.drypipe/*/state.* 2>/dev/null || true"
            stdout = self.invoke_remote(cmd)

            return [
                TaskState(f)
                for f in stdout.strip().split("\n")
                if f != ""
            ]

    def fetch_logs_and_history(self, task):

        remote_pid_basename = os.path.basename(task.pipeline_instance.pipeline_instance_dir)

        #Not called because too slow !
        def fetch_remote_state():
            cmd = f"find {self.remote_base_dir}/{remote_pid_basename}/.drypipe/{task.key}/state.* 2>/dev/null"

            stdout = self.invoke_remote(cmd)

            return TaskState(stdout.strip())

        with self.ssh_client.open_sftp() as sftp:

            def file_content_and_last_mod_time(p):
                f = os.path.join(self.remote_base_dir, remote_pid_basename, p)
                s = sftp.stat(f)
                with sftp.open(f) as _f:
                    return [_f.read(), s.st_mtime]

            f_out, t_out, \
            f_err, t_err, \
            f_history_file, t_history_file = flatten(map(file_content_and_last_mod_time, [
                task.out_log(),
                task.err_log(),
                task.history_file()
            ]))

            last_activity_time = max([t_err, t_out, t_history_file])

            return f_out.decode("utf-8"), f_err.decode("utf-8"), f_history_file.decode("utf-8"), last_activity_time

    def _remote_control_dir(self, task):
        return os.path.join(
            self.remote_base_dir,
            os.path.basename(task.pipeline_instance.pipeline_instance_dir),
            ".drypipe",
            task.key
        )

    def _file_in_control_dir(self, task, file):
        return os.path.join(
            self._remote_control_dir(task),
            file
        )

    def clear_remote_task_state(self, task):
        state_file = self._file_in_control_dir(task, "state.*")
        self.invoke_remote(f"rm {state_file}", bash_error_ok=True)

    def kill_slurm_task(self, task):

        slurm_job_id_file = self._file_in_control_dir(task, "slurm_job_id")

        with self.ssh_client.open_sftp() as sftp:

            with sftp.open(slurm_job_id_file) as _f:
                slurm_job_id = _f.read().strip().decode("utf-8")
                self.invoke_remote(f"scancel {slurm_job_id}")

    def delete_remote_state_file(self, file):

        self.invoke_remote(f"rm {file}")

    def _rsync_with_args_and_remote_dir(self):

        ssh_args = f"-e 'ssh -q -i {self.key_filename} -o StrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null'"

        timeout = f"--timeout={60 * 2}"

        return (
            f"rsync {ssh_args} {timeout}",
            f"{self.ssh_username}@{self.ssh_host}:{self.remote_base_dir}"
        )

    def _launch_command(self, command, exception_func=lambda stderr_text: stderr_text):

        with PortablePopen(command,shell=True) as p:
            p.wait_and_raise_if_non_zero()

    def do_rsync_container(self, image_path):

        rsync_call, remote_dir = self._rsync_with_args_and_remote_dir()

        bn = os.path.basename(image_path)
        dn = os.path.dirname(image_path)

        self.invoke_remote(f"mkdir -p {self.remote_base_dir}/pipeline_code_dir")

        cmd = f"{rsync_call} -az --partial {image_path} {remote_dir}/pipeline_code_dir"

        self._launch_command(
            cmd,
            lambda err: Exception(f"uploading container {bn} failed:\n{cmd}\n{err}")
        )

    def rsync_remote_code_dir_if_applies(self, pipeline, task_conf):

        remote_pipeline_code_dir = task_conf.remote_pipeline_code_dir

        if remote_pipeline_code_dir is None:
            return

        with perf_logger_timer("RemoteSSH.rsync_remote_code_dir_if_applies") as t:

            rsync_call, remote_dir = self._rsync_with_args_and_remote_dir()

            self.invoke_remote(f"mkdir -p {remote_pipeline_code_dir}")

            cmd = f"{rsync_call} -az --partial " + \
                  f"{pipeline.pipeline_code_dir}/ {self.ssh_username}@{self.ssh_host}:{remote_pipeline_code_dir}"

            self._launch_command(
                cmd,
                lambda err: Exception(f"rsync of remote_pipeline_code_dir failed:\n{cmd}\n{err}")
            )

    def upload_task_inputs(self, task_state):
        with perf_logger_timer("RemoteSSH.upload_task_inputs") as t:

            task_control_dir = task_state.control_dir()

            rsync_call, remote_dir = self._rsync_with_args_and_remote_dir()
            pipeline_instance_dir = os.path.dirname(os.path.dirname(task_control_dir))

            remote_pid_basename = os.path.basename(pipeline_instance_dir)

            self.invoke_remote(f"mkdir -p {self.remote_base_dir}/{remote_pid_basename}")

            cmd = f"{rsync_call} -caRz --partial --recursive --files-from={task_control_dir}/local-deps.txt " + \
                  f"{pipeline_instance_dir} {remote_dir}/{remote_pid_basename}"

            self._launch_command(
                cmd,
                lambda err: Exception(f"uploading of task inputs {task_control_dir} failed:\n{cmd}\n{err}")
            )

            # NO LONGER SUPPORTED
            # if self.rsync_containers and task.container is not None:
            #    self.do_rsync_container(task.container.image_path)

    def download_task_results(self, task_state):
        with perf_logger_timer("RemoteSSH.download_task_results") as t:
            task_control_dir = task_state.control_dir()

            rsync_call, remote_dir = self._rsync_with_args_and_remote_dir()
            pipeline_instance_dir = os.path.dirname(os.path.dirname(task_control_dir))
            remote_pid_basename = os.path.basename(pipeline_instance_dir)

            cmd = f"{rsync_call} --dirs -aR --partial --files-from={task_control_dir}/remote-outputs.txt " + \
                  f"{remote_dir}/{remote_pid_basename} {pipeline_instance_dir}"

            def error_msg(err):
                r = f"{task_control_dir}/remote-outputs.txt"
                return Exception(
                    f"downloading of task results {r} from {self.ssh_host} failed:\n{cmd}\n{err}"
                )

            self._launch_command(cmd, error_msg)

    def fetch_remote_file(self, remote_file, local_file):
        with perf_logger_timer("RemoteSSH.fetch_remote_file") as t:
            self.ssh_client.copy_remote_file(remote_file, local_file)

    def fetch_remote_file_content(self, remote_file):
        with tempfile.NamedTemporaryFile() as tmp:
            self.fetch_remote_file(remote_file, tmp.name)
            with open(tmp.name) as _f:
                return _f.read()

    def upload_file(self, local_file, remote_file):
        with perf_logger_timer("RemoteSSH.upload_file") as t:
            self.ssh_client.copy_file(local_file, remote_file)

    def prepare_remote_instance_directory(self, pipeline_instance, task_conf):

        if pipeline_instance.is_remote_instance_directory_prepared(self.server_connection_key()):
            return

        pipeline_instance_dir = pipeline_instance.pipeline_instance_dir
        remote_pid_basename = os.path.basename(pipeline_instance_dir)

        overrides_file_for_host = os.path.join(
            pipeline_instance.work_dir,
            f"pipeline-env-{self.server_connection_key()}.sh"
        )

        logger_ssh.info(f"will upload overrides: '%s'", overrides_file_for_host)

        remote_pipeline_code_dir = task_conf.remote_pipeline_code_dir
        remote_containers_dir = task_conf.remote_containers_dir

        if remote_containers_dir is None:
            if remote_pipeline_code_dir is not None:
                remote_containers_dir = os.path.join(remote_pipeline_code_dir, "containers")

        def gen_remote_overrides():
            yield "__pipeline_code_dir", remote_pipeline_code_dir
            if remote_containers_dir is not None:
                yield "__containers_dir", remote_containers_dir

        remote_overrides = [
            f"export {k}={v}\n"
            for k, v in gen_remote_overrides()
            if v is not None
        ]

        r_work_dir = os.path.join(self.remote_base_dir, remote_pid_basename, ".drypipe")

        if len(remote_overrides) > 0:
            logger_ssh.info(f"no overrides to upload")

            with open(overrides_file_for_host, "w") as _f:
                _f.write(f"{bash_shebang()}\n\n")
                _f.writelines(remote_overrides)
                _f.write("\n")

            r_override_file = os.path.join(
                r_work_dir,
                "pipeline-env.sh"
            )

            self.upload_file(overrides_file_for_host, r_override_file)

        def local_to_remote(file):
            return os.path.join(pipeline_instance.work_dir, file), os.path.join(r_work_dir, file)

        self.upload_file(*local_to_remote('script_lib.py'))
        script_lib_l, script_lib_r = local_to_remote('script_lib')
        self.upload_file(script_lib_l, script_lib_r)
        self.invoke_remote(f"chmod u+x {script_lib_r}")

        pipeline_instance.set_remote_instance_directory_prepared(self.server_connection_key())

    """
        Fetches log lines and history.txt for all tasks, done once every janitor run, instead of before each task
    """
    def fetch_new_log_lines(self, pipeline_instance_dir):
        with perf_logger_timer("RemoteSSH.fetch_new_log_lines") as t:

            rsync_call, remote_dir = self._rsync_with_args_and_remote_dir()
            remote_pid_basename = os.path.basename(pipeline_instance_dir)
            dst = os.path.dirname(pipeline_instance_dir)

            cmd = " ".join([
                f" {rsync_call} -r --partial --append",
                " --filter='+ .drypipe/*/*.log'",
                " --filter='+ .drypipe/*/history.tsv'",
                " --filter='- .drypipe/*/*'",
                " --filter='- output'",
                " --filter='- *.sh'",
                f" {remote_dir}/{remote_pid_basename} {dst}"
            ])

            logger_ssh.debug("will rsync new log lines \n%s", cmd)

            self._launch_command(
                cmd,
                lambda err: Exception(
                    f"fetching logs failed:\n{cmd}\n{err}")
            )

    def execute(self, task, touch_pid_file_func, wait_for_completion=False, fail_silently=False):

        task_local_logger = create_task_logger(os.path.join(task.v_abs_control_dir()))
        try:

            remote_pid_basename = os.path.basename(task.pipeline_instance.pipeline_instance_dir)

            remote_pipeline_instance_dir = os.path.join(self.remote_base_dir, remote_pid_basename)

            remote_script_lib_path = os.path.join(
                remote_pipeline_instance_dir,
                '.drypipe',
                'script_lib'
            )

            drypipe_task_debug = os.environ.get("DRYPIPE_TASK_DEBUG") == "True"

            cmd = " ".join([
                remote_script_lib_path,
                "launch-task-from-remote",
                task.key,
                "--is-slurm" if task.task_conf.is_slurm() else "",
                "--wait-for-completion" if wait_for_completion else "",
                "--drypipe-task-debug" if drypipe_task_debug else ""
            ])

            task_local_logger.info("will execute")

            task_local_logger.debug(f"will launch task on %s: %s", self.ssh_host, cmd)

            with perf_logger_timer("RemoteSSH.execute") as t:
                res = self.invoke_remote(cmd)
                try:
                    res_num = int(res)
                except Exception as ex:
                    raise Exception(f"remote command {cmd} returned invalid result: {res}")
                if res_num != 0:
                    raise Exception(f"remote command {cmd} failed with return code: {res_num}")
        finally:
            task_local_logger.handlers[0].close()
